#!/usr/bin/python3
"""NRPE check to verify if mongodb replicas are healthy."""

#
# Copyright 2020 Canonical Ltd.
#
# Author: David O Neill <david.o.neill@canonical.com>
#

import json
import subprocess
import sys

import yaml

NAGIOS_OK = 0
NAGIOS_WARN = 1
NAGIOS_CRIT = 2
NAGIOS_UNKNOWN = 3

REPL_STATES_OK = ["PRIMARY", "SECONDARY", "ARBITER"]
# Unused, but for reference:
# REPL_STATES_WARN = ["STARTUP", "STARTUP2", "ROLLBACK", "RECOVERING"]
REPL_STATES_CRIT = ["UNKNOWN", "DOWN", "REMOVED"]


def get_cmd_result(cmd):
    try:
        return subprocess.check_output(cmd).decode("utf8")
    except subprocess.CalledProcessError as e:
        print("CRITICAL: failed executing {}, error was {}".format(cmd, e))
        sys.exit(NAGIOS_CRIT)


def get_replica_set_name():
    # mongo2 file format the key is 'replSet'
    # mongo34 yaml file format the key is 'replSetName'
    # this will match both

    with open("/etc/mongodb.conf", "r") as mongo_file:
        data = mongo_file.read()
        try:
            config = yaml.safe_load(data)
        except yaml.scanner.ScannerError:
            config = None

        if isinstance(config, dict):
            if "replication" not in config:
                return False
            if "replSetName" not in config["replication"]:
                return False
            return config["replication"]["replSetName"]

        for line in data.splitlines():
            if line.strip().startswith("replSet") is False:
                continue
            parts = line.split("=")
            return parts[1].strip()

    return False


def get_replica_status():
    # https://bugs.launchpad.net/charm-mongodb/+bug/1892122
    # Looking for RECOVERING
    # {
    #     "members" : [
    #         {
    #             ...
    #             "stateStr" : "SECONDARY",
    #             ...
    #         },
    #         ...
    #     ]
    # }

    name = get_replica_set_name()
    if not name:
        print("CRITIAL: unable to determine the replica set name")
        sys.exit(NAGIOS_CRIT)

    mongo_res = get_cmd_result(
        ["/usr/bin/mongo", name, "--quiet", "--eval", "JSON.stringify(rs.status())"]
    )

    try:
        res_json = json.loads(mongo_res)
    except json.decoder.JSONDecodeError:
        print("CRITICAL: failed to load json")
        sys.exit(NAGIOS_CRIT)

    warnings = []
    errors = []

    for member in res_json["members"]:
        state = member["stateStr"]
        if state in REPL_STATES_OK:
            continue
        elif state in REPL_STATES_CRIT:
            errors.append("{}: {}".format(member["name"], state))
        else:
            warnings.append("{}: {}".format(member["name"], state))

    if errors:
        print(
            "CRITICAL: the following members are in error states: {}".format(
                ", ".join(errors)
            )
        )
        sys.exit(NAGIOS_CRIT)

    if warnings:
        print(
            "WARN: the following members are in warning states: {}".format(
                ", ".join(warnings)
            )
        )
        sys.exit(NAGIOS_WARN)

    print("OK")
    sys.exit(NAGIOS_OK)


if __name__ == "__main__":
    get_replica_status()
