#!/usr/bin/env python3
"""Created on Aug 1, 2012.

@author: negronjl
"""

import collections
import distutils
import json
import os
import pprint
import re
import signal
import socket
import subprocess
import sys
import time
from os import chmod, remove
from os.path import exists
from string import Template
from textwrap import dedent

try:
    import yaml  # noqa
except ImportError:
    if sys.version_info.major == 2:
        subprocess.check_call(["apt-get", "install", "-y", "python-yaml"])
    else:
        subprocess.check_call(["apt-get", "install", "-y", "python3-yaml"])
    import yaml  # noqa

from charmhelpers.contrib.charmsupport.nrpe import NRPE
from charmhelpers.contrib.hahelpers.cluster import peer_units
from charmhelpers.core.decorators import retry_on_exception
from charmhelpers.core.hookenv import (
    DEBUG,
    Hooks,
    UnregisteredHookError,
    WARNING,
    application_version_set,
    charm_dir,
    close_port,
    config,
    env_proxy_settings,
    is_leader,
    is_relation_made,
    local_unit,
    log as juju_log,
    open_port,
    relation_get,
    relation_id,
    relation_ids,
    relation_set,
    relations_of_type,
    status_set,
    unit_get,
)
from charmhelpers.core.host import (
    file_hash,
    lsb_release,
    rsync,
    service,
)
from charmhelpers.fetch import add_source, apt_install, apt_update
from charmhelpers.fetch.python.packages import pip_install
from charmhelpers.payload.execd import execd_preinstall

from yaml.constructor import ConstructorError

try:
    from pymongo import MongoClient
    from pymongo.errors import OperationFailure
except ImportError:
    if sys.version_info.major == 2:
        apt_install("python-pymongo", fatal=True)
    else:
        apt_install("python3-pymongo", fatal=True)
    from pymongo import MongoClient
    from pymongo.errors import OperationFailure

try:
    import distro  # noqa
except ImportError:
    proxy_settings = env_proxy_settings()
    proxy = None
    if proxy_settings and "https_proxy" in proxy_settings:
        proxy = proxy_settings["https_proxy"]
    pip_install("distro", proxy=proxy)
    import distro  # noqa

try:
    from packaging import version
except ImportError:
    if sys.version_info.major == 2:
        apt_install("python-packaging", fatal=True)
    else:
        apt_install("python3-packaging", fatal=True)
    from packaging import version

hooks = Hooks()

###############################################################################
# Global variables
###############################################################################
percona_mongodb_pid_file = "/var/run/mongod.pid"
percona_mongodb_config = "/etc/mongod.conf"
default_mongodb_config = "/etc/mongodb.conf"
default_mongodb_pid_file = "/var/run/mongodb/mongos.pid"
default_mongodb_init_config = "/etc/init/mongodb.conf"
mongodb_env_config = "/etc/default/mongodb"
default_mongos_list = "/etc/mongos.list"
default_wait_for = 3
default_max_tries = 7

INSTALL_PACKAGES = ["python3-yaml"]

# number of seconds init_replset will pause while looping to check if
# replicaset is initialized
INIT_CHECK_DELAY = 1.5

# number of times mongo_client_smart will try to execute given statement
MONGO_CLIENT_RETRIES = 10

# These are MongoDB ReplicaSet states, for convenience:
MONGO_STARTUP = 0
MONGO_PRIMARY = 1
MONGO_SECONDARY = 2
MONGO_RECOVERING = 3
MONGO_FATAL = 4
MONGO_STARTUP2 = 5
MONGO_UNKNOWN = 6
MONGO_ARBITER = 7
MONGO_DOWN = 8
MONGO_ROLLBACK = 9
MONGO_REMOVED = 10

# Ill and quick way to make relation-departed and relation-broken communicate
was_i_primary = False

###############################################################################
# Supporting functions
###############################################################################


def is_percona_source():
    """Determine if percona repository is used.

    This function will check the source repository of mongodb package
    and return true if it comes from percona repo because it will change the whole charm
    process like the config file structure.
    An example of the percona mongodb repository is:
    http://repo.percona.com/psmdb-44/apt  main
    """
    if "repo.percona.com" in config("source") or config("percona-mode"):
        return True
    else:
        return False


def is_bionic_or_greater():
    current_version = distro.linux_distribution()[1]
    if distutils.version.LooseVersion(
        current_version
    ) >= distutils.version.LooseVersion("18.04"):
        return True


def port_check(host=None, port=None, protocol="TCP"):
    if host is None or port is None:
        juju_log("port_check: host and port must be defined.")
        return False
    if protocol.upper() == "TCP":
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    elif protocol.upper() == "UDP":
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    else:
        juju_log("port_check: Unrecognized protocol %s" % protocol)
        return False
    try:
        s.connect((host, int(port)))
        s.shutdown(socket.SHUT_RDWR)
        juju_log("port_check: %s:%s/%s is open" % (host, port, protocol))
        return True
    except Exception as e:
        juju_log("port_check: Unable to connect to %s:%s/%s." % (host, port, protocol))
        juju_log("port_check: Exception: %s" % str(e))
        return False


# update_service_ports:  Convenience function that evaluate the old and new
#                        service ports to decide which ports need to be
#                        opened and which to close
def update_service_ports(old_service_ports=None, new_service_ports=None):
    juju_log("update_service_ports")
    if old_service_ports is None or new_service_ports is None:
        return None
    for port in old_service_ports:
        if port not in new_service_ports:
            juju_log("closing port: %d" % int(port))
            close_port(port)
    for port in new_service_ports:
        if port not in old_service_ports:
            juju_log("opening port: %d" % int(port))
            open_port(port)


def regex_sub(pat_replace=None, data=None):
    juju_log("regex_sub")
    if not pat_replace or not data:
        raise Exception("pat_replace or data not defined")
    if not isinstance(pat_replace, list):
        raise Exception("pat_replace must be a list of pat, replace tuples")
    new_data = data
    for (pattern, replace) in pat_replace:
        new_data = re.sub(pattern, replace, data, 0, re.MULTILINE)
    return new_data


def update_file(filename=None, new_data=None, old_data=None):
    juju_log("update_file: %s" % filename, level=DEBUG)
    if filename is None or new_data is None:
        retval = False
    try:
        if old_data != new_data:
            with open(filename, "w") as f:
                f.write(new_data)
        retval = True
    except Exception as e:
        juju_log(str(e))
        retval = False
    finally:
        return retval


def process_check(pid=None):
    try:
        if pid is not None:
            cmd_line = subprocess.check_output(
                "ps -p %d -o cmd h" % int(pid), shell=True
            )
            retval = (pid, cmd_line)
        else:
            juju_log("process_check: pid not defined.")
            retval = (None, None)
    except Exception as e:
        juju_log("process_check exception: %s" % str(e))
        retval = (None, None)
    finally:
        juju_log("process_check returs pid: %s and cmd_line: %s" % retval)
        return retval


def process_check_pidfile(pidfile=None):
    if pidfile is not None and os.path.exists(pidfile):
        return process_check(open(pidfile).read())
    else:
        juju_log("process_check_pidfile: undefined or non-existant pidfile.")
        return (None, None)


def is_valid_ip(bind_ip):
    if bind_ip == unit_get("private-address") or bind_ip == unit_get("public-address"):
        return True
    return False


def choose_bind_ip(bind_ip):
    default_value, localhost = "0.0.0.0", "127.0.0.1"
    known_values = collections.defaultdict(lambda: False)
    known_values.update(
        {
            "0.0.0.0": default_value,
            "all": default_value,
            "localhost": localhost,
            "127.0.0.1": localhost,
        }
    )

    if known_values[bind_ip]:
        return known_values[bind_ip]

    if not is_valid_ip(bind_ip):
        juju_log("bind_ip not valid. Using default '0.0.0.0' instead")
        return default_value

    return bind_ip


class MasterNotFoundException(Exception):  # NOQA: N818
    """Define MasterNotFoundException."""

    pass


class TimeoutException(Exception):  # NOQA: N818
    """Define TimeoutException."""

    pass


###############################################################################
# Charm support functions
###############################################################################
def mongodb_conf(config_data=None):  # noqa: C901
    if config_data is None:
        return None
    config = []

    # header
    config.append("# mongodb.conf")
    config.append("")

    # dbpath
    # Create the directory if not there already
    subprocess.call(["mkdir", "-p", "%s" % config_data["dbpath"]])
    # Make sure the mongodb user has access to it
    subprocess.call(["chown", "-R", "mongodb:mongodb", config_data["dbpath"]])
    config.append("dbpath=%s" % config_data["dbpath"])
    config.append("")
    config.append("ipv6=true")

    # logpath
    # Create the directory if not there already
    subprocess.call(["mkdir", "-p", "%s" % os.path.dirname(config_data["logpath"])])
    subprocess.call(
        ["chown", "-R", "mongodb:mongodb", os.path.dirname(config_data["logpath"])]
    )

    config.append("logpath=%s" % config_data["logpath"])
    config.append("")

    # log_append
    if config_data["logappend"]:
        config.append("logappend=true")
        config.append("")

    # bind_ip
    bind_ip = choose_bind_ip(config_data["bind_ip"])
    config.append("bind_ip = %s" % bind_ip)
    config.append("")

    # port
    config.append("port = %d" % config_data["port"])
    config.append("")

    # journal
    if config_data["journal"]:
        config.append("journal=true")
        config.append("")

    # cpu
    if config_data["cpu"]:
        config.append("cpu = true")
        config.append("")

    # auth
    if config_data["auth"]:
        config.append("auth = true")
        config.append("")

    # verbose
    if config_data["verbose"]:
        config.append("verbose = true")
        config.append("")

    # objcheck
    if config_data["objcheck"]:
        config.append("objcheck = true")
        config.append("")

    # quota
    if config_data["quota"]:
        config.append("quota = true")
        config.append("quotaFiles = {}".format(config_data["quotafiles"]))
        config.append("")

    if not is_bionic_or_greater():
        # diaglog
        config.append("diaglog = %d" % config_data["diaglog"])
        config.append("")
        # nohttpinterface
        if config_data["web_admin_ui"]:
            config.append("rest = true")
            config.append("")

    # noscripting
    if config_data["noscripting"]:
        config.append("noscripting = true")
        config.append("")

    # notablescan
    if config_data["notablescan"]:
        config.append("notablescan = true")
        config.append("")

    # noprealloc
    if config_data["noprealloc"]:
        config.append("noprealloc = true")
        config.append("")

    # nssize
    if config_data["nssize"] != "default":
        config.append("nssize = %s" % config_data["nssize"])
        config.append("")

    # Set either replica-set or master, depending upon whether the
    # the replica-set (peer) relation is established. If the user
    # chooses to use juju scale out (e.g. juju add-unit) then the
    # charm will use replica-set replication. The user may opt to
    # do master/slave replication or sharding as a different form
    # of scaleout.
    if is_relation_made("replica-set"):
        config.append("replSet = %s" % config_data["replicaset"])
        config.append("")
    else:
        if config_data["master"] == "self":
            config.append("master = true")
            config.append("")
        else:
            config.append("slave = true")
            config.append("source = %s" % config_data["master"])
            config.append("")

    # arbiter
    if config_data["arbiter"] != "disabled" and config_data["arbiter"] != "enabled":
        config.append("arbiter = %s" % config_data["arbiter"])
        config.append("")

    # autoresync
    if config_data["autoresync"]:
        config.append("autoresync")
        config.append("")

    # oplogSize
    if config_data["oplogSize"] != "default":
        config.append("oplogSize = %s" % config_data["oplogSize"])
        config.append("")

    # extra config options
    if config_data["extra_config_options"] != "none":
        for config_option in config_data["extra_config_options"].split(","):
            config.append(config_option)

    return "\n".join(config)


def setup_directory_structure_and_permissions(user_group, config_data=None):
    # Create the dbpath directory if not there already
    subprocess.call(["mkdir", "-p", config_data["dbpath"]])
    # Make sure the mongodb user has access to it
    subprocess.call(["chown", "-R", user_group, config_data["dbpath"]])
    # Create the logpath directory if not there already
    subprocess.call(["mkdir", "-p", os.path.dirname(config_data["logpath"])])
    subprocess.call(
        ["chown", "-R", user_group, os.path.dirname(config_data["logpath"])]
    )


def mongodb_conf_yaml(config_data=None):  # noqa: C901
    if config_data is None:
        return None
    # https://docs.mongodb.com/v4.2/reference/configuration-options/#file-format
    config = {}
    if is_percona_source():
        mongodb_pid_file = percona_mongodb_pid_file
        user_group = "mongod:mongod"
    else:
        mongodb_pid_file = default_mongodb_pid_file
        user_group = "mongodb:mongodb"

    setup_directory_structure_and_permissions(user_group, config_data)

    config["storage"] = {
        "dbPath": config_data["dbpath"],
    }

    config["systemLog"] = {
        "destination": "file",
        "path": config_data["logpath"],
    }

    # log_append
    if config_data["logappend"]:
        config["systemLog"]["logAppend"] = True

    # bind_ip and port
    config["net"] = {}
    bind_ip = choose_bind_ip(config_data["bind_ip"])
    config["net"]["bindIp"] = bind_ip

    # port
    config["net"]["port"] = config_data["port"]

    # ipv6
    config["net"]["ipv6"] = True

    # journal
    if config_data["journal"]:
        if "storage" not in config:
            config["storage"] = {}
        config["storage"]["journal"] = {}
        config["storage"]["journal"]["enabled"] = True

    # auth
    if config_data["auth"]:
        if "security" not in config:
            config["security"] = {}
        config["security"] = {
            "authorization": "enabled",
        }

    # verbose
    if config_data["verbose"]:
        if "systemLog" not in config:
            config["systemLog"] = {}
        config["systemLog"]["verbosity"] = 5
        config["systemLog"]["quiet"] = False
        config["systemLog"]["traceAllExceptions"] = True

    # objcheck
    if config_data["objcheck"]:
        if "net" not in config:
            config["net"] = {}
        config["net"]["wireObjectCheck"] = True

    # Configure processManagement section
    if is_percona_source():
        process_mgmt_dict = {"fork": "true", "pidFilePath": mongodb_pid_file}
        config["processManagement"] = process_mgmt_dict

    # "cpu", "quota", "rest", "noprealloc", "nssize", "master", "arbiter",
    # "autoresync", "extra_config_options" are deprecated when >= 4.2
    # https://docs.mongodb.com/manual/reference/configuration-options/#removed-mmapv1-options

    # noscripting
    if config_data["noscripting"]:
        if "security" not in config:
            config["security"] = {}
        config["security"]["javascriptEnabled"] = False

    # notablescan
    if config_data["notablescan"]:
        if "setParameter" not in config:
            config["setParameter"] = {}
        config["setParameter"]["notablescan"] = True

    # Set either replica-set or master, depending upon whether the
    # the replica-set (peer) relation is established. If the user
    # chooses to use juju scale out (e.g. juju add-unit) then the
    # charm will use replica-set replication. The user may opt to
    # do master/slave replication or sharding as a different form
    # of scaleout.
    if is_relation_made("replica-set"):
        if "replication" not in config:
            config["replication"] = {}
        config["replication"]["replSetName"] = config_data["replicaset"]

    # oplogSize
    if config_data["oplogSize"] != "default":
        if "replication" not in config:
            config["replication"] = {}
        config["replication"]["oplogSizeMB"] = config_data["oplogSize"]

    # header + dictionary to yaml
    config_str = "# mongodb.conf"
    config_str += "\r\n\r\n"
    config_str += yaml.dump(config)

    return config_str


def get_current_mongo_config():
    """Read the current mongo configuration file and returns a dict.

    :returns dict: key/value pairs of the configuration file.
    """
    results = {}
    if is_percona_source():
        mongodb_config = "/etc/mongod.conf"
    else:
        mongodb_config = default_mongodb_config

    with open(mongodb_config, "r") as f:
        for line in f:
            line = line.strip()

            # Skip over comments, blank lines, and any other line
            # that appears to not contain a key = value pair.
            if line.startswith("#") or "=" not in line:
                continue

            key, value = line.split("=", 1)
            results[key.strip()] = value.strip()

        if not results:
            with open(mongodb_config, "r") as f:
                results = yaml.safe_load(f.read())

    return results


def mongo_client(host=None, command=None):
    if host is None or command is None:
        return False
    else:
        cmd_line = "mongo"
        cmd_line += " --host %s" % host
        cmd_line += " --eval 'printjson(%s)'" % command
        juju_log("Executing: %s" % cmd_line, level=DEBUG)
        return subprocess.call(cmd_line, shell=True) == 0


def mongo_client_smart(host="localhost", command=None):
    """Rework of mongo_client function.

    This function retries the command MONGO_CLIENT_RETRIES times

    :param host: The host to connect to. Defaults to localhost
    :param command: The command to be executed. Can't be None
    :returns True if command succeeded, False if it failed
    """
    if command is None:
        return False

    cmd_line = ["mongo", "--quiet", "--host", host, "--eval", "printjson(%s)" % command]
    juju_log("mongo_client_smart executing: %s" % str(cmd_line), level=DEBUG)

    for i in range(MONGO_CLIENT_RETRIES):
        try:
            cmd_output = subprocess.check_output(cmd_line).decode("utf8")
            juju_log("mongo_client_smart executed, output: %s" % cmd_output)
            if json.loads(cmd_output)["ok"] == 1:
                return True
        except subprocess.CalledProcessError as err:
            juju_log("mongo_client_smart failed: %s" % err.output, level=DEBUG)
            pass
        finally:
            time.sleep(1.5)

    # At this point, the command failed
    juju_log("mongo_client_smart failed executing: %s", level=WARNING)
    return False


def init_replset():
    config_data = config()
    # Use my IP at rs.initiate(), voids issues with invalid (and/or
    # not resolvable by peers) hostnames

    rset = config_data["replicaset"]
    addr = unit_get("private-address")
    port = config_data["port"]

    init = '{_id: "%s", members: [{_id: 0, host: "%s:%s"}]} ' % (rset, addr, port)

    retval = mongo_client("localhost", "rs.initiate(%s)" % init)
    time.sleep(1)  # give mongod some time to become primary
    c = MongoClient("localhost")
    while True:
        try:
            r = run_admin_command(c, "replSetGetStatus")
            mongo_state = r["myState"]
            juju_log("init_replset: myState: %s" % mongo_state)
            if mongo_state == MONGO_PRIMARY:  # we're primary!
                break
            elif mongo_state in (
                MONGO_STARTUP,
                MONGO_STARTUP2,
                MONGO_SECONDARY,
            ):  # we are still initializing
                continue
            else:
                juju_log("init_replset: Unexpected replicaSet state: %s" % mongo_state)
                retval = False
                break
        except OperationFailure as e:
            juju_log("init_replset: OperationFailure: %s" % e, DEBUG)
            if "Received replSetInitiate" in str(e):
                continue
            else:
                juju_log("init_replset: Unhandled OperationFailure %s" % e)
                raise
        finally:
            time.sleep(INIT_CHECK_DELAY)

    juju_log("init_replset returns: %s" % retval, level=DEBUG)
    return retval


def run_admin_command(client, cmdstr):
    """Run an admin command against the client.

    Primary purpose is to simplify the unit testing.
    """
    return client.admin.command(cmdstr)


def join_replset(master_node=None, host=None):
    # TODO: This method shouldn't accept any arguments because we don't care
    #       about master_node - we always run replset admin commands on
    #       localhost.
    #       However, that might break other code calling this method.
    #       This will wait charm rewrite.
    juju_log("join_replset: master_node: %s, host: %s" % (master_node, host))
    if master_node is None or host is None:
        retval = False
    else:
        retval = rs_add(host)
    juju_log("join_replset returns: %s" % retval, level=DEBUG)
    return retval


def leave_replset(master_node=None, host=None):
    juju_log("leave_replset: master_node: %s, host: %s" % (master_node, host))
    if master_node is None or host is None:
        retval = False
    else:
        retval = mongo_client(master_node, 'rs.remove("%s")' % host)
    juju_log("leave_replset returns: %s" % retval, level=DEBUG)
    return retval


def enable_replset(replicaset_name=None):
    if is_percona_source():
        default_mongodb_config = percona_mongodb_config
    retval = False
    if replicaset_name is None:
        juju_log("enable_replset: replicaset_name is None, exiting", level=DEBUG)
    try:
        juju_log("enable_replset: Enabling replicaset configuration:")

        current_config = get_current_mongo_config()
        config_data = dict(config())

        if (
            "replSet" in current_config
            and current_config["replSet"] == config_data["replicaset"]
        ):
            juju_log("enable_replset: replica set is already enabled", level=DEBUG)
        else:
            juju_log(
                "enable_replset: enabling replicaset %s" % config_data["replicaset"],
                level=DEBUG,
            )
            if version.parse(get_mongod_version()) < version.parse("4.2"):
                mongodb_config = mongodb_conf(config_data)
            else:
                mongodb_config = mongodb_conf_yaml(config_data)
            retval = update_file(default_mongodb_config, mongodb_config)

        juju_log("enable_replset will return: %s" % str(retval), level=DEBUG)

    except Exception as e:
        juju_log(str(e), level=WARNING)
        retval = False
    finally:
        return retval


def remove_replset_from_upstart():
    """Remove replicaset configuration from upstart."""
    try:
        with (open(default_mongodb_init_config, "r")) as f:
            mongodb_init_config = f.read()

        if re.search(" --replSet", mongodb_init_config, re.MULTILINE) is not None:
            mongodb_init_config = re.sub(r" --replSet .\w+", "", mongodb_init_config)
        retval = update_file(default_mongodb_init_config, mongodb_init_config)
    except Exception as e:
        juju_log(str(e))
        retval = False
    finally:
        return retval


def step_down_replset_primary():
    """Step down the primary."""
    retval = mongo_client("localhost", "rs.stepDown()")
    for i in range(MONGO_CLIENT_RETRIES):
        if not am_i_primary():
            juju_log("step_down_replset_primary returns: %s" % retval, level=DEBUG)
            return retval
        # If we are still primary wait a bit more for step down
        time.sleep(1.5)

    # Raise an error if we exhausted the maximum amount of trials
    raise TimeoutException("Unable to step down the primary")


def remove_rest_from_upstart():
    """Remove --rest from upstart script."""
    try:
        mongodb_init_config = open(default_mongodb_init_config).read()
        if re.search(" --rest ", mongodb_init_config, re.MULTILINE) is not None:
            mongodb_init_config = regex_sub([(" --rest ", " ")], mongodb_init_config)
        retval = update_file(default_mongodb_init_config, mongodb_init_config)
    except Exception as e:
        juju_log(str(e))
        retval = False
    finally:
        return retval


def update_daemon_options(daemon_options=None):
    if is_bionic_or_greater():
        if daemon_options and daemon_options != "none":
            daemon_opts = 'DAEMON_OPTS="{0}"\n'.format(daemon_options)
            return update_file(mongodb_env_config, daemon_opts)
        else:
            if os.path.exists(mongodb_env_config):
                os.remove(mongodb_env_config)
                return True
    else:
        mongodb_init_config = open(default_mongodb_init_config).read()
        pat_replace = []
        if daemon_options is None or daemon_options == "none":
            pat_replace.append(
                (" --config /etc/mongodb.conf.*", " --config /etc/mongodb.conf; fi")
            )
        else:
            pat_replace.append(
                (
                    " --config /etc/mongodb.conf.*",
                    " --config /etc/mongodb.conf %s; fi" % daemon_options,
                )
            )
        regex_sub(pat_replace, mongodb_init_config)
        return update_file(default_mongodb_init_config, mongodb_init_config)


def enable_arbiter(master_node=None, host=None):
    juju_log("enable_arbiter: master_node: %s, host: %s" % (master_node, host))
    if master_node is None or host is None:
        retval = False
    else:
        retval = mongo_client(master_node, 'rs.addArb("%s")' % host)
    juju_log("enable_arbiter returns: %s" % retval)
    return retval


def configsvr_status(wait_for=default_wait_for, max_tries=default_max_tries):
    config_data = config()
    current_try = 0

    while (
        (process_check_pidfile("/var/run/mongodb/configsvr.pid") != (None, None))
        and not port_check(
            unit_get("private-address"), config_data["config_server_port"]
        )
        and current_try < max_tries
    ):

        juju_log("configsvr_status: Waiting for Config Server to be ready ...")
        time.sleep(wait_for)
        current_try += 1
    retval = (
        (process_check_pidfile("/var/run/mongodb/configsvr.pid") != (None, None))
        == port_check(unit_get("private-address"), config_data["config_server_port"])
        is True
    )
    if retval:
        return process_check_pidfile("/var/run/mongodb/configsvr.pid")
    else:
        return (None, None)


def configsvr_ready(wait_for=default_wait_for, max_tries=default_max_tries):
    return configsvr_status(wait_for, max_tries) != (None, None)


def disable_configsvr(port=None):
    if port is None:
        juju_log("disable_configsvr: port not defined.")
        return False
    try:
        config_server_port = config("config_server_port")
        pid = open("/var/run/mongodb/configsvr.pid").read()
        os.kill(int(pid), signal.SIGTERM)
        os.unlink("/var/run/mongodb/configsvr.pid")
        retval = True
    except Exception as e:
        juju_log("no config server running ...")
        juju_log("Exception: %s" % str(e))
        retval = False
    finally:
        juju_log("disable_configsvr returns %s" % retval)
        close_port(config_server_port)
        return retval


def enable_configsvr(
    config_data, wait_for=default_wait_for, max_tries=default_max_tries
):
    if config_data is None:
        juju_log("enable_configsvr: config_data not defined.")
        return False

    # Stop any running config servers
    disable_configsvr(config_data["config_server_port"])

    # Make sure dbpath and logpath exist
    subprocess.call(["mkdir", "-p", "%s" % config_data["config_server_dbpath"]])
    subprocess.call(
        ["mkdir", "-p", "%s" % os.path.dirname(config_data["config_server_logpath"])]
    )

    # Start the config server
    juju_log("enable_configsvr: Starting the config server")
    cmd_line = "mongod"
    cmd_line += " --configsvr"
    cmd_line += " --bind_ip {}".format(choose_bind_ip(config_data["bind_ip"]))
    cmd_line += " --port %d" % config_data["config_server_port"]
    cmd_line += " --dbpath %s" % config_data["config_server_dbpath"]
    cmd_line += " --logpath %s" % config_data["config_server_logpath"]
    cmd_line += " --pidfilepath /var/run/mongodb/configsvr.pid"
    cmd_line += " --replSet {}".format(config_data["replicaset"])
    cmd_line += " --fork"
    subprocess.call(cmd_line, shell=True)

    retval = configsvr_ready(wait_for, max_tries)
    if retval:
        open_port(config_data["config_server_port"])
        if config_data["web_admin_ui"]:
            port = int(config_data["config_server_port"]) + 1000
            open_port(port)
    juju_log("enable_configsvr returns: %s" % retval)
    return retval


def mongos_status(wait_for=default_wait_for, max_tries=default_max_tries):
    config_data = config()
    if is_percona_source():
        mongodb_pid_file = percona_mongodb_pid_file
    else:
        mongodb_pid_file = default_mongodb_pid_file
    current_try = 0
    while (
        (process_check_pidfile(mongodb_pid_file) != (None, None))
        and not port_check(unit_get("private-address"), config_data["mongos_port"])
        and current_try < max_tries
    ):

        juju_log("mongos_status: Waiting for Mongo shell to be ready ...")
        time.sleep(wait_for)
        current_try += 1
    retval = (
        (process_check_pidfile(mongodb_pid_file) != (None, None))
        == port_check(unit_get("private-address"), config_data["mongos_port"])
        is True
    )

    if retval:
        return process_check_pidfile(mongodb_pid_file)
    else:
        return (None, None)


def mongos_ready(wait_for=default_wait_for, max_tries=default_max_tries):
    return mongos_status(wait_for, max_tries) != (None, None)


def disable_mongos(port=None):
    if is_percona_source():
        mongodb_pid_file = percona_mongodb_pid_file
    else:
        mongodb_pid_file = default_mongodb_pid_file
    if port is None:
        juju_log("disable_mongos: port not defined")
        return False
    try:
        pid = open(mongodb_pid_file).read()
        os.kill(int(pid), signal.SIGTERM)
        os.unlink(mongodb_pid_file)
        retval = True
    except Exception as e:
        juju_log("no mongo router running ...")
        juju_log("Exception: %s" % str(e))
        retval = False
    finally:
        juju_log("disable_mongos returns %s" % retval)
        close_port(port)
        return retval


def enable_mongos(
    config_data=None,
    config_servers=None,
    replicaset=None,
    wait_for=default_wait_for,
    max_tries=default_max_tries,
):
    if is_percona_source():
        mongodb_pid_file = percona_mongodb_pid_file
    else:
        mongodb_pid_file = default_mongodb_pid_file
    juju_log("enable_mongos")
    if config_data is None or config_servers is None:
        juju_log("enable_mongos: config_data and config_servers are mandatory")
        return False
    if not isinstance(config_servers, list):
        juju_log("enable_mongos: config_servers must be a list")
        return False
    if len(config_servers) < 3:
        # MongoDB 3.2 deprecates the use of three mirrored mongod instances
        # for config servers.
        juju_log("enable_mongos: Not enough config servers yet...")
        return True
    disable_mongos()
    # Make sure logpath exist
    subprocess.call(
        ["mkdir", "-p", "%s" % os.path.dirname(config_data["mongos_logpath"])]
    )
    cmd_line = "mongos"
    cmd_line += " --logpath %s" % config_data["mongos_logpath"]
    cmd_line += " --pidfilepath " + mongodb_pid_file
    cmd_line += " --port %d" % config_data["mongos_port"]
    cmd_line += " --fork"
    # Note(aluria): --configdb used to have a list of comma-separated
    # server:port values. Such list now needs to be prepended by
    # repSetConfigName/
    cmd_line += " --configdb {}/{}".format(replicaset, ",".join(config_servers[0:3]))
    juju_log("enable_mongos: cmd_line: %s" % cmd_line)
    subprocess.call(cmd_line, shell=True)
    retval = mongos_ready(wait_for, max_tries)
    if retval:
        open_port(config_data["mongos_port"])
    juju_log("enable_mongos returns: %s" % retval)
    return retval


def load_config_servers(mongos_list=None):
    if os.path.exists(mongos_list):
        retval = [line.strip() for line in open(mongos_list).readlines()]
    else:
        retval = []
    return retval


def restart_mongod(wait_for=default_wait_for, max_tries=default_max_tries):
    my_hostname = unit_get("private-address")
    my_port = config("port")
    current_try = 0
    service_name = "mongodb"
    if is_percona_source():
        service_name = "mongod"
    service("stop", service_name)
    if os.path.exists("/var/lib/mongodb/mongod.lock"):
        os.remove("/var/lib/mongodb/mongod.lock")

    if not service("start", service_name):
        return False

    while (
        service("status", service_name)
        and not port_check(my_hostname, my_port)
        and current_try < max_tries
    ):
        juju_log(
            "restart_mongod: Waiting for MongoDB to be ready ({}/{})".format(
                current_try, max_tries
            )
        )
        time.sleep(wait_for)
        current_try += 1

    return (service("status", service_name) == port_check(my_hostname, my_port)) is True


def backup_cronjob(disable=False):
    """Generate the cronjob to backup with mongodump."""
    juju_log("Setting up cronjob")
    config_data = config()
    backupdir = config_data["backup_directory"]
    bind_ip = config_data["bind_ip"]
    cron_file = "/etc/cron.d/mongodb"
    cron_runtime = "@daily"

    # Disable or not remove it and regenerate it with new config data.
    if exists(cron_file):
        remove(cron_file)

    if not disable:
        tpl_data = {
            "backup_copies": config_data["backup_copies_kept"],
            "backup_directory": backupdir,
            "unique_name": os.environ.get("JUJU_UNIT_NAME", bind_ip),
            "bind_ip": bind_ip,
            "port": config_data["port"],
        }

        script_filename = "/var/lib/mongodb/cronbackup.py"
        script_template = "templates/backup.py.tpl"

        juju_log("Writing out cronbackup.py")
        with open(script_template) as handle:
            template = Template(handle.read())
            rendered = template.substitute(tpl_data)

            with open(script_filename, "w") as output:
                output.writelines(rendered)
            chmod(script_filename, 0o755)

        juju_log("Installing cron.d/mongodb")

        if exists(cron_file):
            remove(cron_file)

        with open(cron_file, "w") as output:
            output.write(
                """
SHELL=/bin/bash
PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin

%s    ubuntu python %s

"""
                % (cron_runtime, script_filename)
            )


###############################################################################
# Hook functions
###############################################################################
@hooks.hook("install.real")
def install_hook():
    juju_log("Begin install hook.")
    execd_preinstall()
    status_set("maintenance", "Installing packages")
    juju_log("Installing mongodb")
    add_source(config("source"), config("key"))

    apt_update(fatal=True)
    apt_install(packages=INSTALL_PACKAGES, fatal=True)
    apt_install(packages=config("package"), fatal=True)


@hooks.hook("config-changed")  # noqa: C901
def config_changed():  # noqa: C901
    juju_log("Entering config_changed")
    status_set("maintenance", "Configuring unit")
    config_data = config()
    juju_log("config_data: {}".format(config_data), level=DEBUG)

    if is_percona_source():
        with open(percona_mongodb_config) as file:
            mongodb_config = yaml.safe_load(file)
        mongodb_config_hash = file_hash(percona_mongodb_config)
    else:
        mongodb_config = open(default_mongodb_config).read()
        mongodb_config_hash = file_hash(default_mongodb_config)
    if (
        config_data.changed("volume-ephemeral-storage")
        or config_data.changed("volume-map")
        or config_data.changed("volume-dev-regexp")
    ):
        config_changed_volume()

    # current ports
    if is_percona_source():
        current_mongodb_port = mongodb_config["net"]["port"]
    else:
        current_mongodb_port = re.search(
            r"^#*port\s+=\s+(\w+)", mongodb_config, re.MULTILINE
        ).group(1)

    current_web_admin_ui_port = int(current_mongodb_port) + 1000
    new_web_admin_ui_port = int(config_data["port"]) + 1000

    juju_log("Configured mongodb port: {}".format(current_mongodb_port), level=DEBUG)
    public_address = unit_get("public-address")
    juju_log("unit's public_address: {}".format(public_address), level=DEBUG)
    private_address = unit_get("private-address")
    juju_log("unit's private_address: {}".format(private_address), level=DEBUG)

    # Update mongodb configuration file
    if version.parse(get_mongod_version()) < version.parse("4.2"):
        mongodb_config = mongodb_conf(config_data)
    else:
        mongodb_config = mongodb_conf_yaml(config_data)
    if is_percona_source():
        update_file(percona_mongodb_config, mongodb_config)
    else:
        update_file(default_mongodb_config, mongodb_config)

    if config_data["backups_enabled"]:
        backup_cronjob()
    else:
        backup_cronjob(disable=True)

    # web_admin_ui
    if config_data["web_admin_ui"]:
        open_port(new_web_admin_ui_port)
    else:
        close_port(current_web_admin_ui_port)

    # replicaset_master
    if config_data["replicaset_master"] != "auto":
        enable_replset(config_data["replicaset"])
        join_replset(config_data["replicaset_master"])

    # extra demon options
    update_daemon_options(config_data["extra_daemon_options"])

    # write mongodb logrotate configuration file
    write_logrotate_config(config_data)

    # restart mongodb if the configuration file has changed.
    if file_hash(default_mongodb_config) != mongodb_config_hash:
        status_set("maintenance", "Restarting mongod")
        restart_mongod()

    # attach to replSet ( if needed )
    if config_data["replicaset_master"] != "auto":
        join_replset(config_data["replicaset_master"], private_address)

    # arbiter
    if config_data["replicaset_master"] != "auto":
        if (
            config_data["arbiter"] != "disabled"
            and config_data["replicaset_master"] != "auto"
        ):
            if config_data["arbiter"] == "enable":
                enable_arbiter(
                    config_data["replicaset_master"],
                    "%s:%s" % (private_address, config_data["port"]),
                )
            else:
                enable_arbiter(config_data["replicaset_master"], config_data["arbiter"])

    # expose necessary ports
    update_service_ports([current_mongodb_port], [config_data["port"]])

    if config_data["web_admin_ui"]:
        current_web_admin_ui_port = int(current_mongodb_port) + 1000
        new_web_admin_ui_port = int(config_data["port"]) + 1000
        close_port(current_web_admin_ui_port)
        open_port(new_web_admin_ui_port)

    # update config-server information and port
    try:
        (configsvr_pid, configsvr_cmd_line) = configsvr_status()
    except Exception as e:
        configsvr_pid = None
        configsvr_cmd_line = None
        juju_log("config_changed: configsvr_status failed.")
        juju_log("config_changed: Exception: %s" % str(e))

    if configsvr_pid is not None:
        configsvr_port = re.search(r"--port (\w+)", configsvr_cmd_line).group(2)
        disable_configsvr(configsvr_port)
        enable_configsvr(config_data["config_server_port"])
    else:
        open_port(config_data["config_server_port"])

    # update mongos information and port
    try:
        (mongos_pid, mongos_cmd_line) = mongos_status()
    except Exception as e:
        mongos_pid = None
        mongos_cmd_line = None
        juju_log("config_changed: mongos_status failed.")
        juju_log("config_changed: Exceptions: %s" % str(e))

    if mongos_pid is not None:
        mongos_port = re.search(r"--port (\w+)", mongos_cmd_line).group(1)
        disable_mongos(mongos_port)
        enable_mongos(config_data["mongos_port"])
    else:
        open_port(config_data["mongos_port"])

    update_nrpe_config()
    application_version_set(get_mongod_version())
    update_status()

    juju_log("About to leave config_changed", level=DEBUG)
    return True


@hooks.hook("start")
def start_hook():
    juju_log("start_hook")
    retval = restart_mongod()
    juju_log("start_hook returns: %s" % retval)
    return retval


@hooks.hook("stop")
def stop_hook():
    juju_log("stop_hook")
    try:
        retval = service("stop", "mongodb")
        os.remove("/var/lib/mongodb/mongod.lock")
        # FIXME Need to check if this is still needed
    except Exception as e:
        juju_log(str(e))
        retval = False
    finally:
        juju_log("stop_hook returns: %s" % retval)
        return retval


@hooks.hook("database-relation-joined")
def database_relation_joined():
    juju_log("database_relation_joined")
    my_hostname = unit_get("private-address")
    my_port = config("port")
    my_replset = config("replicaset")
    juju_log("my_hostname: %s" % my_hostname)
    juju_log("my_port: %s" % my_port)
    juju_log("my_replset: %s" % my_replset)
    relation_data = {
        "hostname": my_hostname,
        "port": my_port,
        "type": "database",
        "version": get_mongod_version(),
    }

    if len(peer_units("replica-set")) > 1:
        relation_data["replset"] = my_replset

    relation_set(relation_id(), relation_data)


@hooks.hook("replica-set-relation-joined")
def replica_set_relation_joined():
    juju_log("replica_set_relation_joined-start")
    my_hostname = unit_get("private-address")
    my_port = config("port")
    my_replset = config("replicaset")

    my_install_order = os.environ["JUJU_UNIT_NAME"].split("/")[1]
    juju_log("my_hostname: %s" % my_hostname)
    juju_log("my_port: %s" % my_port)
    juju_log("my_replset: %s" % my_replset)
    juju_log("my_install_order: %s" % my_install_order)
    # do not restart mongodb if config has not changed
    if enable_replset(my_replset):
        juju_log("Restarting mongodb after config change (enable replset)", level=DEBUG)
        status_set("maintenance", "Restarting mongod to enable replicaset")
        restart_mongod()

    relation_set(
        relation_id(),
        {
            "hostname": my_hostname,
            "port": my_port,
            "replset": my_replset,
            "install-order": my_install_order,
            "type": "replset",
        },
    )

    update_status()
    juju_log("replica_set_relation_joined-finish")


def rs_add(host):
    if not is_bionic_or_greater():
        return mongo_client_smart("localhost", 'rs.add("%s")' % host)

    command = 'rs.add("%s")' % host
    if host is None:
        raise ValueError("missing host")
    else:
        cmd_line = [
            "mongo",
            "--quiet",
            "--host",
            "localhost",
            "--eval",
            "printjson(%s)" % command,
        ]
        juju_log("Executing: %s" % cmd_line, level=DEBUG)
        run(cmd_line)

    for i in range(MONGO_CLIENT_RETRIES):
        c = MongoClient("localhost")
        subprocess.check_output(cmd_line)
        r = run_admin_command(c, "replSetGetStatus")
        members = r["members"]
        ok = [m for m in members if m["name"] == host and m["state"] == MONGO_SECONDARY]
        if ok:
            return ok

        time.sleep(1.5)

    return False


def am_i_primary():
    c = MongoClient("localhost")
    for i in range(10):
        try:
            r = run_admin_command(c, "replSetGetStatus")
            pretty_r = pprint.pformat(r)
            juju_log(
                "am_i_primary: replSetGetStatus returned: %s" % pretty_r, level=DEBUG
            )
            return r["myState"] == MONGO_PRIMARY
        except OperationFailure as e:
            juju_log("am_i_primary: OperationError: %s" % str(e), level=DEBUG)
            if "replSetInitiate - should come online shortly" in str(e):
                # replSet initialization in progress
                continue
            elif "EMPTYCONFIG" in str(e):
                # replication not initialized yet
                return False
            elif "no replset config has been received" in str(e):
                # replication not initialized yet (Mongo3.4+)
                return False
            elif "not running with --replSet" in str(e):
                # replicaset not configured
                return False
            else:
                raise
        finally:
            time.sleep(1.5)

    # Raise an error if we exhausted the maximum amount of trials
    raise TimeoutException("Unable to determine if local unit is primary")


def get_replicaset_status():
    """Connect to mongod and get the status of replicaset.

    This function is used mainly within update_status() to display
    replicaset status in 'juju status' output

    :returns string: can be any of replicaset states
        (https://docs.mongodb.com/manual/reference/replica-states/)
        or can be the string of an exception while getting the status
    """
    c = MongoClient("localhost")
    try:
        r = run_admin_command(c, "replSetGetStatus")
        for member in r["members"]:
            if "self" in member:
                return member["stateStr"]

        # if 'self' was not found in the output, then log a warning and print
        # the output given by replSetGetStatus
        r_pretty = pprint.pformat(r)
        juju_log(
            "get_replicaset_status() failed to get replicaset state: %s" % r_pretty,
            level=WARNING,
        )
        return "Unknown replica set state"

    except OperationFailure as e:
        juju_log("get_replicaset_status() exception: %s" % str(e), DEBUG)
        if "not running with --replSet" in str(e):
            return "not in replicaset"
        else:
            return str(e)


def get_mongod_version():
    """Connect to mongod and get the db.version() output.

    Mainly used for application_set_version in config-changed hook
    """
    c = MongoClient("localhost", serverSelectionTimeoutMS=60000)
    return c.server_info()["version"]


# Retry until the replica set is in active state,
# retry 45 times before failing,
# wait 1, 2, 3, 4, ... seconds between each retry,
# this will add 33 minutes of accumulated sleep()
@retry_on_exception(num_retries=45, base_delay=1)
def wait_until_replset_is_active():
    status = update_status()
    if status != "active":
        raise Exception("ReplicaSet not active: {}".format(status))


@hooks.hook("replica-set-relation-changed")
def replica_set_relation_changed():
    private_address = unit_get("private-address")
    remote_hostname = relation_get("hostname")

    juju_log("replica_set_relation_changed-start")
    juju_log(
        "local unit: %s, joining_unit: %s" % (private_address, remote_hostname),
        level=DEBUG,
    )

    if remote_hostname is None:
        juju_log("Joiner not ready yet... bailing out")
        return

    # Initialize the replicaset - we do this only on the leader!
    if is_leader():
        juju_log("Initializing replicaset")
        status_set("maintenance", "Initializing replicaset")
        init_replset()

    unit = "%s:%s" % (private_address, config("port"))
    unit_remote = "%s:%s" % (remote_hostname, relation_get("port"))

    # If this is primary, add joined unit to replicaset
    if am_i_primary():
        juju_log("Adding new secondary... %s" % unit_remote, level=DEBUG)
        join_replset(unit, unit_remote)

    wait_until_replset_is_active()
    juju_log("replica_set_relation_changed-finish")


@hooks.hook("replica-set-relation-departed")
def replica_set_relation_departed():
    juju_log("replica_set_relation_departed-start")

    if not am_i_primary():
        juju_log("replica_set_relation_departed-finish")
        return

    unit_address, unit_port = unit_get("private-address"), config("port")
    remote_address = relation_get("private-address")
    remote_port = relation_get("port")

    # If I am the unit being removed, step me down from being primary
    if (unit_address, unit_port) == (remote_address, remote_port):
        juju_log("Stepping down from being primary...")
        global was_i_primary
        was_i_primary = True
        mongo_client("localhost", "rs.stepDown()")
        juju_log("replica_set_relation_departed-finish")
        return

    unit = "%s:%s" % (unit_get("private-address"), config("port"))
    unit_remote = "%s:%s" % (relation_get("hostname"), relation_get("port"))

    leave_replset(unit, unit_remote)
    juju_log("Removed %s from replicaset" % unit_remote)
    juju_log("replica_set_relation_departed-finish")


@hooks.hook("replica-set-relation-broken")
def replica_set_relation_broken():
    juju_log("replica_set_relation_broken-start")

    if am_i_primary():
        juju_log("I was primary - removing myself via new primary.", "DEBUG")
        mongo_client("localhost", "rs.stepDown()")
        time.sleep(15)  # give some time to for re-election to happen

    c = MongoClient("localhost")
    r = c.admin.command("isMaster")

    try:
        master_node = r["primary"]
    except KeyError:
        pass

    if "master_node" in locals():  # unit is part of replicaset, remove it!
        unit = "%s:%s" % (unit_get("private-address"), config("port"))
        juju_log("Removing myself via %s" % (master_node), "DEBUG")
        leave_replset(master_node, unit)

    juju_log("replica_set_relation_broken-finish")


@hooks.hook("data-relation-joined")
def data_relation_joined():
    juju_log("data_relation_joined")

    return relation_set(relation_id(), {"mountpoint": "/srv/juju/mongodb-data"})


@hooks.hook("data-relation-changed")
def data_relation_changed():
    juju_log("data_relation_changed")

    if volume_get_id_for_storage_subordinate() is None:
        juju_log("mountpoint from storage subordinate not ready, let's wait")
        return True

    config_changed_volume()


@hooks.hook("data-relation-departed")
def data_relation_departed():
    juju_log("data_relation_departed")
    return config_changed_volume()


@hooks.hook("configsvr-relation-joined")
def configsvr_relation_joined():
    juju_log("configsvr_relation_joined")
    my_hostname = unit_get("private-address")
    my_port = config("config_server_port")
    my_install_order = os.environ["JUJU_UNIT_NAME"].split("/")[1]
    my_replicaset = config("replicaset")
    relation_set(
        relation_id(),
        {
            "hostname": my_hostname,
            "port": my_port,
            "install-order": my_install_order,
            "replset": my_replicaset,
            "type": "configsvr",
        },
    )


@hooks.hook("configsvr-relation-changed")
def configsvr_relation_changed():
    juju_log("configsvr_relation_changed")
    config_data = config()
    my_port = config_data["config_server_port"]
    disable_configsvr(my_port)
    retval = enable_configsvr(config_data)
    juju_log("configsvr_relation_changed returns: %s" % retval)
    return retval


@hooks.hook("mongos-cfg-relation-joined")
@hooks.hook("mongos-relation-joined")
def mongos_relation_joined():
    juju_log("mongos_relation_joined")
    my_hostname = unit_get("private-address")
    my_port = config("mongos_port")
    my_install_order = os.environ["JUJU_UNIT_NAME"].split("/")[1]
    my_replicaset = config("replicaset")
    relation_set(
        relation_id(),
        {
            "hostname": my_hostname,
            "port": my_port,
            "install-order": my_install_order,
            "replset": my_replicaset,
            "type": "mongos",
        },
    )


@hooks.hook("mongos-cfg-relation-changed")
@hooks.hook("mongos-relation-changed")
def mongos_relation_changed():
    juju_log("mongos_relation_changed")
    config_data = config()
    retval = False

    hostname = relation_get("hostname")
    port = relation_get("port")
    replicaset = relation_get("replset")
    rel_type = relation_get("type")
    if hostname is None or port is None or rel_type is None:
        juju_log("mongos_relation_changed: relation data not ready.", level=DEBUG)
        return
    if rel_type == "configsvr":
        config_servers = load_config_servers(default_mongos_list)
        juju_log("Adding config server: %s:%s" % (hostname, port), level=DEBUG)
        if hostname and port and "{}:{}".format(hostname, port) not in config_servers:
            config_servers.append("%s:%s" % (hostname, port))
        disable_mongos(config_data["mongos_port"])
        retval = enable_mongos(config_data, config_servers, replicaset)
        if retval:
            update_file(default_mongos_list, "\n".join(config_servers))
    elif rel_type == "database":
        if mongos_ready():
            mongos_host = "%s:%s" % (unit_get("private-address"), config("mongos_port"))
            shard_command1 = 'sh.addShard("%s:%s")' % (hostname, port)
            mongo_client(mongos_host, shard_command1)
            replicaset = relation_get("replset")
            shard_command2 = 'sh.addShard("%s/%s:%s")' % (replicaset, hostname, port)
            mongo_client(mongos_host, shard_command2)

    else:
        juju_log(
            "mongos_relation_change: undefined rel_type: %s" % rel_type, level=DEBUG
        )
        return

    juju_log("mongos_relation_changed returns: %s" % retval, level=DEBUG)


@hooks.hook("mongos-relation-broken")
def mongos_relation_broken():
    config_servers = load_config_servers(default_mongos_list)
    for member in relations_of_type("mongos"):
        hostname = relation_get("hostname", member)
        port = relation_get("port", member)
        if "%s:%s" % (hostname, port) in config_servers:
            config_servers.remove("%s:%s" % (hostname, port))

    update_file(default_mongos_list, "\n".join(config_servers))


@hooks.hook(
    "nrpe-external-master-relation-joined", "nrpe-external-master-relation-changed"
)
def update_nrpe_config():
    # Find out if nrpe set nagios_hostname
    hostname = None
    host_context = None
    reldata = relations_of_type("nrpe-external-master")
    if not reldata:
        return
    for rel in reldata:
        if "nagios_hostname" in rel:
            hostname = rel["nagios_hostname"]
            host_context = rel["nagios_host_context"]
            break
    nrpe = NRPE(hostname=hostname)
    apt_install("python3-dbus")

    if host_context:
        current_unit = "%s:%s" % (host_context, local_unit())
    else:
        current_unit = local_unit()

    if lsb_release()["DISTRIB_RELEASE"] > "15.04":
        check_mongo_script = "check_systemd.py mongodb"
    else:
        check_mongo_script = "check_upstart_job mongodb"

    nrpe.add_check(
        shortname="mongodb",
        description="process check {%s}" % current_unit,
        check_cmd=check_mongo_script,
    )

    charm_plugin_dir = os.path.join(charm_dir(), "files", "nrpe-external-master/")
    rsync(
        charm_plugin_dir, "/usr/local/lib/nagios/plugins/", options=["--executability"]
    )

    nrpe.remove_check(shortname="check_replica_sets")
    nrpe.remove_check(shortname="replica_sets_status")
    if is_relation_made("replica-set"):
        nrpe.add_check(
            shortname="replica_sets_status",
            description="replica set check {%s}" % current_unit,
            check_cmd="check_replica_sets.py",
        )

    nrpe.write()


@hooks.hook("upgrade-charm")
def uprade_charm():
    juju_log("upgrade-charm: removing --replset from upstart script")
    remove_replset_from_upstart()
    juju_log("upgrade-charm: removing --rest from upstart script")
    remove_rest_from_upstart()


@hooks.hook("update-status")
def update_status():
    """Run update status hook.

    This hook both updates status and provides workload_state return value
    for other functions to determine if update_status is needed
    Returns: workload_state
    """
    workload = "active"
    status = "Unit is ready"

    if is_relation_made("replica-set"):
        # only check for replica-set state if the relation was made which means
        # more than 1 units were deployed and peer related.
        mongo_status = get_replicaset_status()
        if mongo_status in ("PRIMARY", "SECONDARY"):
            workload = "active"
            status = "Unit is ready as " + mongo_status
        elif mongo_status in ("not in replicaset"):
            workload = "active"
            status = "Unit is ready, " + mongo_status
        else:
            workload = "maintenance"
            status = mongo_status
            juju_log("mongo_status is unknown: {}".format(status), level=DEBUG)

    juju_log("Setting workload: {} - {}".format(workload, status), level=DEBUG)
    status_set(workload, status)

    return workload


def run(command, exit_on_error=True):
    """Run a command and return the output."""
    try:
        juju_log(command)
        return subprocess.check_output(command, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        juju_log("status=%d, output=%s" % (e.returncode, e.output))
        if exit_on_error:
            sys.exit(e.returncode)
        else:
            raise


###############################################################################
# Volume managment
###############################################################################
#
# Get volume-id from juju config "volume-map" dictionary as
#     volume-map[JUJU_UNIT_NAME]
# @return  volid
#
def volume_get_volid_from_volume_map():
    config_data = config()
    volume_map = {}
    try:
        volume_map = yaml.load(config_data["volume-map"].strip())
        if volume_map:
            juju_unit_name = os.environ["JUJU_UNIT_NAME"]
            volid = volume_map.get(juju_unit_name)
            juju_log("Juju unit name: %s Volid:%s" % (juju_unit_name, volid))
            return volid
    except ConstructorError as e:
        juju_log("invalid YAML in 'volume-map': {}".format(e))
    return None


# Is this volume_id permanent ?
# @returns  True if volid set and not --ephemeral, else:
#           False
def volume_is_permanent(volid):
    if volid and volid != "--ephemeral":
        return True
    return False


# Returns a mount point from passed vol-id, e.g. /srv/juju/vol-000012345
#
# @param  volid          volume id (as e.g. EBS volid)
# @return mntpoint_path  eg /srv/juju/vol-000012345
def volume_mount_point_from_volid(volid):
    if volid and volume_is_permanent(volid):
        return "/srv/juju/%s" % volid
    return None


# Returns a stub volume id based on the mountpoint from the storage
# subordinate relation, if present.
#
# @return volid  eg vol-000012345
def volume_get_id_for_storage_subordinate():
    # storage charm is a subordinate so we should only ever have one
    # relation_id for the data relation
    ids = relation_ids("data")
    if len(ids) > 0:
        mountpoint = relation_get("mountpoint", os.environ["JUJU_UNIT_NAME"], ids[0])

        juju_log("mountpoint: %s" % (mountpoint))
        if mountpoint and os.path.exists(mountpoint):
            return mountpoint.split("/")[-1]


# Do we have a valid storage state?
# @returns  volid
#           None    config state is invalid - we should not serve
def volume_get_volume_id():

    config_data = config()
    volid = volume_get_id_for_storage_subordinate()
    if volid:
        return volid

    ephemeral_storage = config_data["volume-ephemeral-storage"]
    volid = volume_get_volid_from_volume_map()
    juju_unit_name = os.environ["JUJU_UNIT_NAME"]
    if ephemeral_storage in [True, "yes", "Yes", "true", "True"]:
        if volid:
            juju_log(
                "volume-ephemeral-storage is True, but"
                "volume-map[{!r}] -> {}".format(juju_unit_name, volid)
            )
            return None
        else:
            return "--ephemeral"
    else:
        if not volid:
            juju_log(
                "volume-ephemeral-storage is False, but "
                "no volid found for volume-map[{!r}]".format(juju_unit_name)
            )
            return None
        juju_log("Volid:%s" % (volid))
    return volid


# Re-initialize the mounted data volume, if necessary.
def config_changed_volume():
    # Trigger volume initialization logic for permanent storage
    volid = volume_get_volume_id()
    if not volid:
        # Invalid configuration (whether ephemeral, or permanent)
        stop_hook()
        mounts = volume_get_all_mounted()
        if mounts:
            juju_log("current mounted volumes: {}".format(mounts))
        juju_log(
            "Disabled and stopped mongodb service, "
            "because of broken volume configuration - check "
            "'volume-ephemeral-storage' and 'volume-map'"
        )
        sys.exit(1)
    if volume_is_permanent(volid):
        # config_changed_volume_apply will stop the service if it finds
        # it necessary, ie: new volume setup
        if config_changed_volume_apply():
            start_hook()
        else:
            stop_hook()
            mounts = volume_get_all_mounted()
            if mounts:
                juju_log("current mounted volumes: {}".format(mounts))
            juju_log(
                "Disabled and stopped mongodb service "
                "(config_changed_volume_apply failure)"
            )
            sys.exit(1)


# Initialize and/or mount permanent storage, it straightly calls
# shell helper
def volume_init_and_mount(volid):
    juju_log("Initialize and mount volume")
    command = "scripts/volume-common.sh call volume_init_and_mount %s" % volid
    run(command)
    return True


def volume_get_all_mounted():
    command = "mount |egrep /srv/juju"
    status, output = subprocess.getstatusoutput(command)
    if status != 0:
        return None
    return output


#
# Core logic for permanent storage changes:
# NOTE the only 2 "True" return points:
#   1) symlink already pointing to existing storage (no-op)
#   2) new storage properly initialized:
#     - volume: initialized if not already (fdisk, mkfs),
#       mounts it to e.g.:  /srv/juju/vol-000012345
#     - if fresh new storage dir: rsync existing data
#     - manipulate /var/lib/mongodb/VERSION/CLUSTER symlink
#
def config_changed_volume_apply():  # noqa: C901 is too complex (12)
    config_data = config()
    data_directory_path = config_data["dbpath"]
    assert data_directory_path
    volid = volume_get_volume_id()
    if volid:
        volid_from_subordinate = volume_get_id_for_storage_subordinate()
        if volume_is_permanent(volid) and not volid_from_subordinate:
            if not volume_init_and_mount(volid):
                juju_log("volume_init_and_mount failed, not applying changes")
                return False

        if not os.path.exists(data_directory_path):
            juju_log(
                "mongodb data dir {} not found, "
                "not applying changes.".format(data_directory_path)
            )
            return False

        mount_point = volume_mount_point_from_volid(volid)
        new_mongo_dir = os.path.join(mount_point, "mongodb")
        if not mount_point:
            juju_log(
                "invalid mount point from volid = {}, "
                "not applying changes.".format(mount_point)
            )
            return False

        if os.path.islink(data_directory_path):
            juju_log(
                "mongodb data dir '%s' already points "
                "to %s, skipping storage changes."
                % (data_directory_path, new_mongo_dir)
            )
            juju_log(
                "existing-symlink: to fix/avoid UID changes from "
                "previous units, doing: "
                "chown -R mongodb:mongodb {}".format(new_mongo_dir)
            )
            run("chown -R mongodb:mongodb %s" % new_mongo_dir)
            return True

        # Create a directory structure below "new" mount_point
        curr_dir_stat = os.stat(data_directory_path)
        new_mongo_dir_just_created = False
        if not os.path.isdir(new_mongo_dir):
            juju_log("mkdir %s" % new_mongo_dir)
            os.mkdir(new_mongo_dir)
            new_mongo_dir_just_created = True
            # copy permissions from current data_directory_path
            os.chown(new_mongo_dir, curr_dir_stat.st_uid, curr_dir_stat.st_gid)
            os.chmod(new_mongo_dir, curr_dir_stat.st_mode)
        # Carefully build this symlink, e.g.:
        # /var/lib/mongodb ->
        # /srv/juju/vol-000012345/mongodb
        # but keep previous "main/"  directory, by renaming it to
        # main-$TIMESTAMP
        if not stop_hook():
            juju_log("stop_hook() failed - can't migrate data.")
            return False
        if not os.path.exists(new_mongo_dir) or new_mongo_dir_just_created:
            juju_log(
                "migrating mongo data {}/ -> {}/".format(
                    data_directory_path, new_mongo_dir
                )
            )
            # void copying PID file to perm storage (shouldn't be any...)
            command = "rsync -a {}/ {}/".format(data_directory_path, new_mongo_dir)
            juju_log("run: {}".format(command))
            run(command)
        try:
            os.rename(
                data_directory_path,
                "{}-{}".format(data_directory_path, int(time.time())),
            )
            juju_log(
                "NOTICE: symlinking {} -> {}".format(new_mongo_dir, data_directory_path)
            )
            os.symlink(new_mongo_dir, data_directory_path)
            juju_log(
                "after-symlink: to fix/avoid UID changes from "
                "previous units, doing: "
                "chown -R mongodb:mongodb {}".format(new_mongo_dir)
            )
            run("chown -R mongodb:mongodb {}".format(new_mongo_dir))
            return True
        except OSError:
            juju_log(
                "failed to symlink {} -> {}".format(data_directory_path, mount_point)
            )
            return False
    else:
        juju_log("Invalid volume storage configuration, not applying changes")
    return False


# Write mongodb-server logrotate configuration
def write_logrotate_config(config_data, conf_file="/etc/logrotate.d/mongodb-server"):

    juju_log("Writing {}.".format(conf_file))
    contents = dedent(
        """
        {logpath} {{
                {logrotate-frequency}
                rotate {logrotate-rotate}
                maxsize {logrotate-maxsize}
                copytruncate
                delaycompress
                compress
                notifempty
                missingok
        }}"""
    )
    contents = contents.format(**config_data)
    try:
        with open(conf_file, "w") as f:
            f.write(contents)
    except IOError:
        juju_log("Could not write {}.".format(conf_file))
        return False
    return True


@hooks.hook("pre-series-upgrade")
def pre_series_upgrade():
    juju_log("Running prepare series upgrade hook", "INFO")
    if am_i_primary():
        step_down_replset_primary()
    service("stop", "mongodb")
    status_set(
        "blocked",
        "Ready for do-release-upgrade and reboot. Set complete when finished.",
    )


@hooks.hook("post-series-upgrade")
def post_series_upgrade():
    juju_log("Running complete series upgrade hook", "INFO")
    update_status()


###############################################################################
# Main section
###############################################################################
if __name__ == "__main__":
    # execute a hook based on the name the program is called by
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as e:
        juju_log("Unknown hook {} - skipping".format(e))
