#!/usr/bin/env python3
"""Test class for mongodb zaza tests."""
import unittest

from pymongo import MongoClient

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from zaza import model
from zaza.charm_lifecycle import utils as lifecycle_utils


MONGO_STARTUP = 0
MONGO_PRIMARY = 1
MONGO_SECONDARY = 2
MONGO_RECOVERING = 3
MONGO_FATAL = 4
MONGO_STARTUP2 = 5
MONGO_UNKNOWN = 6
MONGO_ARBITER = 7
MONGO_DOWN = 8
MONGO_ROLLBACK = 9
MONGO_REMOVED = 10


def requests_retry_session(
    retries=3, backoff_factor=2, status_forcelist=(500, 502, 504), session=None
):
    """Create a http session with retry."""
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


class MongodbCharmTestBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.model_name = model.get_juju_model()
        cls.test_config = lifecycle_utils.get_charm_config()
        model.block_until_all_units_idle()
        addr = model.get_lead_unit_ip("mongodb")
        if ":" in addr:  # ipv6 formatting
            cls.leader_address = "[{}]".format(addr)
        else:
            cls.leader_address = addr
        cls.db_client = MongoClient(cls.leader_address)

    def cat_unit(self, unit, path):
        unit_res = model.run_on_unit(unit, "sudo cat {}".format(path))
        return unit_res["Stdout"]

    def web_admin_interface(self, ipaddr, port=28017):
        url = "http://{}:{}".format(ipaddr, port)
        resp = requests_retry_session(retries=10).get(url)
        return resp


class BasicMongodbCharmTest(MongodbCharmTestBase):
    def test_db_insert(self):
        """Test if we can insert and remove a value."""
        test_db = self.db_client.test_db
        insert_id = test_db.testcoll.insert({"assert": True})
        self.assertTrue(insert_id is not None, "Failed to insert test data")
        result = test_db.testcoll.remove(insert_id)
        self.assertTrue("err" not in result, "Failed to remove test data")

    def test_service_running(self):
        """Test if we have a mongod running on all units."""
        for unit in model.get_units("mongodb"):
            running_for = model.get_unit_service_start_time(
                unit.entity_id, "mongod", timeout=20
            )
            self.assertGreater(running_for, 0)


class XenialMongodbCharmTest(MongodbCharmTestBase):
    def test_status_interface(self):
        """Check if we can access the web admin port -- xenial only."""
        resp = self.web_admin_interface(self.leader_address)
        resp.raise_for_status()


class ReplicatedMongodbCharmTest(MongodbCharmTestBase):
    def get_set_status(self):
        unit_status = []
        for addr in model.get_app_ips("mongodb"):
            unit_client = MongoClient(addr)
            unit_status.append(unit_client.admin.command("replSetGetStatus"))
        return unit_status

    def test_replset_numbers(self):
        """Test if we have 1 primary and 2 secondary mongodbs."""
        unit_status = self.get_set_status()
        primaries = [u for u in unit_status if u["myState"] == MONGO_PRIMARY]
        secondaries = [u for u in unit_status if u["myState"] == MONGO_SECONDARY]
        self.assertEqual(len(primaries), 1)
        self.assertEqual(len(secondaries), 2)

    def test_replset_consistent_members(self):
        """Test if all units have the same view on membership."""
        unit_status = self.get_set_status()
        prim_members = [u for u in unit_status if u["myState"] == MONGO_PRIMARY][0][
            "members"
        ]
        secondary_members = [
            u["members"] for u in unit_status if u["myState"] == MONGO_SECONDARY
        ]

        def extract(member_dict):
            # Extract a subset of membership info as a frozenset.
            # Name is ipaddr:port, health a float and stateStr a str
            return frozenset(
                v for k, v in member_dict.items() if k in ["name", "health", "stateStr"]
            )

        ref = set(
            map(extract, prim_members)
        )  # Our reference view on membership comes from the primary
        secondaries = [set(map(extract, sec)) for sec in secondary_members]
        for sec in secondaries:
            self.assertEqual(ref, sec)


class ShardedMongodbCharmTest(MongodbCharmTestBase):
    def test_mongos_running(self):
        """Test if the mongos service is running."""
        running_for = model.get_unit_service_start_time(
            "mongodb/0", "mongos", timeout=20
        )
        self.assertGreater(running_for, 0)


class CharmOperationTest(MongodbCharmTestBase):
    """Verify charm operations."""

    def test_01_nrpe_mongodb_check(self):
        """Verify mongodb is running."""
        expected_nrpe_check = (
            "command[check_mongodb]=/usr/local/lib/nagios/plugins/"
            "check_systemd.py mongodb"
        )
        cmd = "cat /etc/nagios/nrpe.d/check_mongodb.cfg"
        for unit in model.get_units("mongodb"):
            result = model.run_on_unit(unit.entity_id, cmd)
            code = result.get("Code")
            if code != "0":
                raise model.CommandRunFailed(cmd, result)
            content = result.get("Stdout")
            self.assertTrue(expected_nrpe_check in content)

    def test_02_nrpe_check_replica_sets_status(self):
        """Verify that the NRPE replica sets status check is configured correctly."""
        expected_nrpe_check = (
            "command[check_replica_sets_status]=/usr/local/lib/nagios/"
            "plugins/check_replica_sets.py"
        )
        remote_file = "/etc/nagios/nrpe.d/check_replica_sets_status.cfg"
        cmd = "cat {}".format(remote_file)
        units = model.get_units("mongodb")
        if len(units) == 1:
            # Testing a single mongodb node deployment
            result = model.run_on_unit(units[0].entity_id, cmd)
            code = result.get("Code")
            # We expect this command to fail
            if code != "1":
                raise model.CommandRunFailed(cmd, result)
            content = result.get("Stderr")
            print(content)
            self.assertTrue("No such file or directory" in content)
        else:
            # Testing a replica set mongodb deployment
            model.block_until_file_has_contents(
                "mongodb", remote_file, expected_nrpe_check
            )
