"""Test check_replica_sets.py nrpe check."""
import unittest
import unittest.mock as mock

from check_replica_sets import get_cmd_result, get_replica_set_name, get_replica_status

import yaml

try:
    from StringIO import StringIO  # for Python 2
except ImportError:
    from io import StringIO  # for Python 3


class TestCheckReplicaSets(unittest.TestCase):
    """Test class."""

    @mock.patch("check_replica_sets.subprocess.check_output", autospec=True)
    def test_get_cmd_result_true(self, mock_check_output):
        """Test result of subprocess call."""
        mock_check_output.return_value = b"1"
        output = get_cmd_result("echo 1")
        self.assertEqual(output, "1")

    @mock.patch("check_replica_sets.subprocess.check_output", autospec=True)
    def test_get_cmd_result_false(self, mock_check_output):
        """Test error cmd via subprocess call."""
        result = "CRITICAL: failed executing {cmd}, error was {error}".format(
            cmd="dont care", error="Boom!"
        )
        with mock.patch("sys.stdout", new=StringIO()) as fake_std_out:
            try:
                get_cmd_result("dont care")
            except Exception:
                self.assertEqual(fake_std_out.getvalue(), result)

    def test_get_replica_set_name_yaml_replsetname(self):
        """Test mongodb.conf file parsing replSetName success."""
        sample = {}
        sample["replication"] = {}
        sample["replication"]["replSetName"] = "test"
        sample_yaml = yaml.dump(sample)

        with mock.patch(
            "builtins.open", mock.mock_open(read_data=sample_yaml)
        ) as mock_file:
            res = get_replica_set_name()
            mock_file.assert_called_with("/etc/mongodb.conf", "r")
            self.assertEqual(res, "test")

    def test_get_replica_set_name_yaml_no_replsetname(self):
        """Test mongodb.conf file parsing fail, where there's no replSetName."""
        sample = {}
        sample["replication"] = {}
        sample_yaml = yaml.dump(sample)

        with mock.patch(
            "builtins.open", mock.mock_open(read_data=sample_yaml)
        ) as mock_file:
            res = get_replica_set_name()
            mock_file.assert_called_with("/etc/mongodb.conf", "r")
            self.assertEqual(res, False)

    def test_get_replica_set_name_yaml_no_replication(self):
        """Test mongodb.conf file parsing fail, with no replication block."""
        sample = {}
        sample["somethingelse"] = {}
        sample_yaml = yaml.dump(sample)

        with mock.patch(
            "builtins.open", mock.mock_open(read_data=sample_yaml)
        ) as mock_file:
            res = get_replica_set_name()
            mock_file.assert_called_with("/etc/mongodb.conf", "r")
            self.assertEqual(res, False)

    def test_get_replica_set_name_ini_true(self):
        """Test mongodb.conf file parsing success in ini format."""
        sample = """
            # mongodb.conf
            dbpath\t=\t/var/lib/mongodb # tab added to make it fail when parsing as YAML
            ipv6=true
            logpath=/var/log/mongodb/mongodb.log
            logappend=true
            port = 27017
            journal=true
            diaglog = 0
            rest = true
            replSet = test
        """

        with mock.patch("builtins.open", mock.mock_open(read_data=sample)) as mock_file:
            res = get_replica_set_name()
            mock_file.assert_called_with("/etc/mongodb.conf", "r")
            self.assertEqual(res, "test")

    def test_get_replica_set_name_ini_no_replset(self):
        """Test mongodb.conf file missing replSet in ini format."""
        sample = """
            # mongodb.conf
            dbpath=/var/lib/mongodb
            ipv6=true
            logpath=/var/log/mongodb/mongodb.log
            logappend=true
            port = 27017
            journal=true
            diaglog = 0
            rest = true
        """

        with mock.patch("builtins.open", mock.mock_open(read_data=sample)) as mock_file:
            res = get_replica_set_name()
            mock_file.assert_called_with("/etc/mongodb.conf", "r")
            self.assertEqual(res, False)

    @mock.patch("check_replica_sets.get_replica_set_name", autospec=True)
    @mock.patch("check_replica_sets.get_cmd_result", autospec=True)
    def test_get_replica_status_ok(
        self, mock_get_cmd_result, mock_get_replica_set_name
    ):
        """Test replica status in known good states gives OK."""
        mock_get_replica_set_name.return_value = "test"
        mock_get_cmd_result.return_value = """{
            "set" : "myset",
            "members" : [
                {
                    "stateStr" : "SECONDARY"
                },
                {
                    "stateStr" : "PRIMARY"
                },
                {
                    "stateStr" : "SECONDARY"
                }
            ]
        }"""

        result = "OK\n"
        with mock.patch("sys.stdout", new=StringIO()) as fake_std_out:
            try:
                get_replica_status()
            except BaseException as error:
                self.assertEqual(fake_std_out.getvalue(), result)
                self.assertEqual("0", str(error))

    @mock.patch("check_replica_sets.get_replica_set_name", autospec=True)
    @mock.patch("check_replica_sets.get_cmd_result", autospec=True)
    def test_get_replica_status_recovering_warn(
        self, mock_get_cmd_result, mock_get_replica_set_name
    ):
        """Test replica status in RECOVERING state gives WARN."""
        mock_get_replica_set_name.return_value = "test"
        mock_get_cmd_result.return_value = """{
            "set" : "myset",
            "members" : [
                {
                    "name": "one",
                    "stateStr" : "SECONDARY"
                },
                {
                    "name": "two",
                    "stateStr" : "RECOVERING"
                },
                {
                    "name": "three",
                    "stateStr" : "RECOVERING"
                }
            ]
        }"""

        result = (
            "WARN: the following members are in warning states: "
            "two: RECOVERING, three: RECOVERING\n"
        )
        with mock.patch("sys.stdout", new=StringIO()) as fake_std_out:
            try:
                get_replica_status()
            except BaseException as error:
                self.assertEqual(fake_std_out.getvalue(), result)
                self.assertEqual("1", str(error))

    @mock.patch("check_replica_sets.get_replica_set_name", autospec=True)
    @mock.patch("check_replica_sets.get_cmd_result", autospec=True)
    def test_get_replica_status_down_crit(
        self, mock_get_cmd_result, mock_get_replica_set_name
    ):
        """Test replica status in DOWN state gives CRIT."""
        mock_get_replica_set_name.return_value = "test"
        mock_get_cmd_result.return_value = """{
            "set" : "myset",
            "members" : [
                {
                    "name": "one",
                    "stateStr" : "SECONDARY"
                },
                {
                    "name": "two",
                    "stateStr" : "PRIMARY"
                },
                {
                    "name": "three",
                    "stateStr" : "DOWN"
                }
            ]
        }"""

        result = "CRITICAL: the following members are in error states: " "three: DOWN\n"
        with mock.patch("sys.stdout", new=StringIO()) as fake_std_out:
            try:
                get_replica_status()
            except BaseException as error:
                self.assertEqual(fake_std_out.getvalue(), result)
                self.assertEqual("2", str(error))

    @mock.patch("check_replica_sets.get_replica_set_name", autospec=True)
    @mock.patch("check_replica_sets.get_cmd_result", autospec=True)
    def test_get_replica_status_no_replicasetname_crit(
        self, mock_get_cmd_result, mock_get_replica_set_name
    ):
        """Test rs.status() failure 1."""
        mock_get_replica_set_name.return_value = False

        result = "CRITIAL: unable to determine the replica set name\n"
        with mock.patch("sys.stdout", new=StringIO()) as fake_std_out:
            try:
                get_replica_status()
            except BaseException as error:
                self.assertEqual(fake_std_out.getvalue(), result)
                self.assertEqual("2", str(error))

    @mock.patch("check_replica_sets.get_replica_set_name", autospec=True)
    @mock.patch("check_replica_sets.get_cmd_result", autospec=True)
    def test_get_replica_status_bad_json_crit(
        self, mock_get_cmd_result, mock_get_replica_set_name
    ):
        """Test rs.status() failure 2."""
        mock_get_replica_set_name.return_value = "dont care"
        mock_get_cmd_result.return_value = "{ this json is bogus }"

        result = "CRITICAL: failed to load json\n"
        with mock.patch("sys.stdout", new=StringIO()) as fake_std_out:
            try:
                get_replica_status()
            except BaseException as error:
                self.assertEqual(fake_std_out.getvalue(), result)
                self.assertEqual("2", str(error))
