"""Test the utils.py module."""
import io
import logging
import os
import unittest
from contextlib import contextmanager

from mock import patch

import yaml


@contextmanager
def mock_open(filename, contents=None):
    """Slightly simpler mock of open to return contents for filename."""

    def mock_file(*args):
        if args[0] == filename:
            return io.StringIO(contents)
        else:
            return open(*args)

    with patch("builtins.open", mock_file):
        yield


def load_config():
    """Load configuration from file.

    Walk backwards from __file__ looking for config.yaml,
    loads and returns the 'options' section'
    """
    config = None
    f = __file__
    while config is None:
        d = os.path.dirname(f)
        if os.path.isfile(os.path.join(d, "config.yaml")):
            config = os.path.join(d, "config.yaml")
            break
        f = d

    if not config:
        logging.error("Could not find config.yaml in any parent directory of %s." % f)
        raise Exception

    with (open(config, "r")) as cfg:
        value = yaml.safe_load(cfg.read())["options"]
    return value


def get_default_config():
    """Load default charm config from config.yaml.

    If no default is set in config.yaml, the returned value is None.
    return as a dict.
    """
    default_config = {}
    config = load_config()
    for k, v in config.items():
        if "default" in v:
            default_config[k] = v["default"]
        else:
            default_config[k] = None
    return default_config


class CharmTestCase(unittest.TestCase):
    def setUp(self, obj, patches):
        super(CharmTestCase, self).setUp()
        self.patches = patches
        self.obj = obj
        self.test_config = TestConfig()
        self.test_relation = TestRelation()
        self.patch_all()

    def patch(self, method):
        _m = patch.object(self.obj, method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    def patch_all(self):
        for method in self.patches:
            setattr(self, method, self.patch(method))


class TestConfig(object):
    def __init__(self):
        """Initialize TestConfig object."""
        self.config = get_default_config()

    def get(self, attr):
        try:
            return self.config[attr]
        except KeyError:
            return None

    def get_all(self):
        return self.config

    def set(self, attr, value):
        if attr not in self.config:
            raise KeyError
        self.config[attr] = value


class TestRelation(object):
    def __init__(self, relation_data={}):
        """Initialize TestRelation object."""
        self.relation_data = relation_data

    def set(self, relation_data):
        self.relation_data = relation_data

    def get(self, attr=None, unit=None, rid=None):
        if attr is None:
            return self.relation_data
        elif attr in self.relation_data:
            return self.relation_data.get(attr)
        return None
